import { SchemaType, SQLDataType, LockedSchemaType } from "./data-types";
import { SQLManyRelation } from "./data-types/many-relation";
import { SQLMeta } from "./data-types/meta";
import { SQLObject } from "./data-types/object";
import { SQLOneRelation } from "./data-types/one-relation";
import { LockedSchema } from "./locked-schema";
import { Schema } from "./schema";
import { SQLFunction, SQLMatch } from "./utils";
export declare const OPERATORS: {
    EQ: string;
    NE: string;
    GT: string;
    ST: string;
    GE: string;
    SE: string;
    NG: string;
    NS: string;
    LK: string;
    NL: string;
    RG: string;
    IN: string;
    NI: string;
    BT: string;
    NB: string;
};
export declare type FilterFlags<Base, Condition> = {
    [Key in keyof Base]: Base[Key] extends Condition ? never : Key;
};
export declare type AllowedKeys<Base, Condition> = FilterFlags<Base, Condition>[keyof Base];
export declare type SQL_QUERY = {
    sql: string;
    values: any[];
};
export declare type StraightCondition<T extends LockedSchema<any>> = {
    [key in keyof SchemaType<LockedSchemaType<T>>]?: {
        [OP in keyof typeof OPERATORS]?: OP extends "IN" | "NI" | "BT" | "NB" ? Array<SQLDataType<SchemaType<LockedSchemaType<T>>[key]> extends SQLOneRelation<any> | SQLManyRelation<any> | SQLMeta<any> ? never : SQLDataType<SchemaType<LockedSchemaType<T>>[key]> extends SQLObject<any, any, any> ? string : SQLDataType<SchemaType<LockedSchemaType<T>>[key]>> : SQLDataType<SchemaType<LockedSchemaType<T>>[key]> extends SQLOneRelation<any> | SQLManyRelation<any> | SQLMeta<any> ? never : SQLDataType<SchemaType<LockedSchemaType<T>>[key]> extends SQLObject<any, any, any> ? string | Array<string> : SQLDataType<SchemaType<LockedSchemaType<T>>[key]> | Array<SQLDataType<SchemaType<LockedSchemaType<T>>[key]>> | null;
    } | (SQLDataType<SchemaType<LockedSchemaType<T>>[key]> extends SQLOneRelation<any> | SQLManyRelation<any> | SQLMeta<any> ? never : SchemaType<LockedSchemaType<T>>[key] extends SQLObject<any, any, any> ? string | Array<string> : SQLDataType<SchemaType<LockedSchemaType<T>>[key]> | Array<SQLDataType<SchemaType<LockedSchemaType<T>>[key]>> | null);
} & {
    _id?: number | {
        [OP in keyof typeof OPERATORS]?: OP extends "IN" | "NI" | "BT" | "NB" ? Array<number> : number;
    } | Array<number>;
};
export declare type MultipleCondition<S extends LockedSchema<any>> = Array<StraightCondition<S> | SQLMatch<LockedSchemaType<S>>>;
export declare type SelectionColumns<S extends LockedSchema<any>> = AllowedKeys<SchemaType<LockedSchemaType<S>>, SQLOneRelation<any> | SQLManyRelation<any> | SQLMeta<any>>;
export declare type ConditionWhere<S extends LockedSchema<any>> = StraightCondition<S> | MultipleCondition<S> | SQLMatch<LockedSchemaType<S>>;
export interface CONDITIONS<S extends LockedSchema<any>> {
    WHERE?: ConditionWhere<S>;
    HAVING?: ConditionWhere<S>;
    GROUP?: Array<SelectionColumns<S> | "_id" | SQLFunction>;
    ORDER?: {
        COLUMN?: SelectionColumns<S> | "_id" | SQLFunction;
        SORT?: "ASC" | "DESC";
    };
    LIMIT?: {
        OFFSET?: number;
        NUMBER?: number;
    };
}
export declare class ConditionsBuilder<S extends Schema<any>> {
    WHERE?: ConditionWhere<LockedSchema<S>>;
    HAVING?: ConditionWhere<LockedSchema<S>>;
    GROUP?: Array<SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction>;
    ORDER?: {
        COLUMN?: SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction;
        SORT?: "ASC" | "DESC";
    };
    LIMIT?: {
        OFFSET?: number;
        NUMBER?: number;
    };
    protected trimLeft: (string: string, charlist: string) => string;
    protected trimRight: (string: string, charlist: string) => string;
    protected trim: (string: string, charlist: string) => string;
    protected forEach: <O extends Record<any, any>>(Object: O, callback: (value: O[keyof O], key: keyof O, Object: O) => any) => any[];
    protected processCondition: (conditions: StraightCondition<LockedSchema<S>> | SQLMatch<S>) => SQL_QUERY;
    protected processConditions: (conditions: MultipleCondition<LockedSchema<S>>) => SQL_QUERY;
    protected processGroupClause: (group: Array<SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction>) => string;
    protected processOrderClause: (order: {
        COLUMN?: SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction;
        SORT?: "ASC" | "DESC";
    }) => string;
    protected processLimitClause: (limit: {
        OFFSET?: number;
        NUMBER?: number;
    }) => string;
    protected processQueryConditions: (conditions: CONDITIONS<LockedSchema<S>>, enableSelectFeatureClauses?: boolean) => SQL_QUERY;
    conditions: (conditions?: CONDITIONS<LockedSchema<S>> | undefined) => this;
    reset: () => this;
    render: (enableSelectFeatureClauses?: boolean) => SQL_QUERY;
    toString: () => string;
}
