"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConnectionManager = void 0;
const exceptions_1 = require("./exceptions");
class ConnectionManager {
    constructor(Id, Connector, Schemas) {
        this.Id = Id;
        this.Connector = Connector;
        this.Schemas = Schemas;
        this.Ready = false;
        this.query = async (sql, values = []) => {
            if (this.Connector.getConnectorOptions().logs)
                console.log("SQL Query Executed::", sql
                    .split("?")
                    .map((part, index) => `${part}${values[index] !== undefined
                    ? `${typeof values[index] === "string"
                        ? `'${values[index]}'`
                        : values[index]}`
                    : ""}`)
                    .join(""), "\n");
            return await this.Connection.query(sql, values);
        };
        this.SchemaToString = (Schema) => Schema.Render()
            .map((SchemaType) => {
            var _a;
            return `CREATE TABLE IF NOT EXISTS \`${SchemaType.name}\`
        (${SchemaType.schema
                .physicalColumnNames()
                .map((columnName) => `\`${columnName}\` ${SchemaType.schema.getColumn(columnName)}`)}) COLLATE=${((_a = this.Connector.getConnectionOptions()) === null || _a === void 0 ? void 0 : _a.charset) ||
                "UTF8MB4_UNICODE_CI"} `;
        })
            .join(";") + ";";
        this.SchemasBuild = () => this.query(this.Schemas.map((schema) => this.SchemaToString(schema)).join(""));
        this.Sync = (Schema) => new Promise(async (resolve, reject) => {
            // Make sure that a database has been selected in the connection.
            if (typeof this.Connector.getConnectionOptions().database !== "string")
                return reject("No database has been selected!");
            else
                try {
                    // Collect Queries
                    const Queries = [];
                    // Process Each Type Of Schema
                    await Promise.all(Schema.Render().map(async (SchemaType) => {
                        // Get All Columns
                        const Whitelist = SchemaType.schema.physicalColumnNames();
                        const DroppedIndexes = [];
                        // Get current columns of schema
                        const CurrentColumns = (await this.query(`SELECT \`COLUMN_NAME\` FROM INFORMATION_SCHEMA.COLUMNS WHERE \`TABLE_SCHEMA\`='${this.Connector.getConnectionOptions().database}' AND \`TABLE_NAME\`='${SchemaType.name}';`))[0]
                            .map(({ COLUMN_NAME }) => COLUMN_NAME)
                            .filter((v) => v);
                        // Collect Blacklist Columns
                        const Blacklist = CurrentColumns.filter((columnName) => !Whitelist.includes(columnName));
                        // Collect Create/Alter Queries
                        let TempColumnName = undefined;
                        // Execute Columns In Series
                        await Whitelist.reduce((promise, columnName) => promise.then((_) => new Promise(async (resolve) => {
                            if (columnName !== "_id") {
                                // Fetch & Remove All Indexes
                                try {
                                    const Results = await this.query(`SHOW INDEX FROM \`${SchemaType.name}\` WHERE \`COLUMN_NAME\`='${columnName}';`);
                                    if (Results &&
                                        Results instanceof Array &&
                                        Results[0] instanceof Array &&
                                        Results[0].length) {
                                        // Indexes Found!
                                        Results[0].map((index) => {
                                            if (!DroppedIndexes.includes(index.Key_name)) {
                                                Queries.push(`ALTER TABLE \`${index.Table}\` DROP INDEX \`${index.Key_name}\``);
                                                DroppedIndexes.push(index.Key_name);
                                            }
                                            return;
                                        });
                                    }
                                }
                                catch (error) {
                                    console.warn("EPIC SQL WARNING:: Error occured while retriving column indexes! Cannot delete previous column indexes. Please manually delete extra indexes.", "\n", error.stack);
                                }
                                // Update Columns
                                if (!CurrentColumns.includes(columnName))
                                    Queries.push(`ALTER TABLE \`${SchemaType.name}\` ADD COLUMN \`${columnName}\` ${SchemaType.schema.getColumn(columnName)}${TempColumnName
                                        ? ` AFTER \`${TempColumnName}\``
                                        : " FIRST"}`);
                                else
                                    Queries.push(`ALTER TABLE \`${SchemaType.name}\` CHANGE COLUMN \`${columnName}\` \`${columnName}\` ${SchemaType.schema.getColumn(columnName)}${TempColumnName
                                        ? ` AFTER \`${TempColumnName}\``
                                        : " FIRST"}`);
                            }
                            TempColumnName = columnName;
                            return resolve();
                        })), Promise.resolve());
                        if (SchemaType.schema.Indexes.length) {
                            // Create Indexes
                            SchemaType.schema.Indexes.forEach((Index) => Queries.push(`ALTER TABLE \`${SchemaType.name}\` ADD ${Index.index} \`${Index.name}\`${Index.columns.length
                                ? ` (\`${Index.columns.join("`, `")}\`)`
                                : ""}`));
                        }
                        // Collect Drop Queries
                        if (Whitelist.length)
                            Blacklist.forEach((columnName) => Queries.push(`ALTER TABLE \`${SchemaType.name}\` DROP COLUMN \`${columnName}\``));
                        else
                            Queries.push(`DROP TABLE IF EXISTS \`${SchemaType.name}\`;`);
                    }));
                    // Execute Query List
                    if (Queries.length)
                        await this.query(Queries.join(";") + ";");
                }
                catch (error) {
                    return reject(error.message || error);
                }
            // Complete Process
            return resolve(undefined);
        });
        this.getId = () => this.Id;
        this.getConnector = () => this.Connector;
        this.end = () => {
            // End Connection
            this.getConnector().end();
            // Remove all Connection Manager Instances on all Schemas
            this.Schemas.forEach((schema) => schema.deleteConnection(this.Id));
            return this;
        };
        this.init = async () => {
            // Check if Already Initiated
            if (this.Ready)
                throw new exceptions_1.SQLException(`Connection Manager has already been initiated!`);
            // Connect to the database
            const Connector = await this.Connector.connect();
            // Get database connection
            this.Connection = Connector.getConnection();
            // Better Transaction Error Handling
            let TransactionBegin = false;
            try {
                if (this.Schemas.length && this.Connector.getConnectorOptions().sync) {
                    // Start Transaction
                    await this.query("BEGIN;");
                    // Transaction has been started
                    TransactionBegin = true;
                    // Create All Required Tables
                    await this.SchemasBuild();
                    // Remove Foreign Key Checks
                    await this.query(`SET FOREIGN_KEY_CHECKS = 0;`);
                    // Syncronize All Schema Columns
                    await Promise.all(this.Schemas.map((schema) => this.Sync(schema)));
                    // Add Back Foreign Key Checks
                    await this.query(`SET FOREIGN_KEY_CHECKS = 1;`);
                    // Commit Transaction
                    await this.query("COMMIT;");
                }
            }
            catch (error) {
                // Rollback Transaction If Transaction Begin
                if (TransactionBegin)
                    await this.query("ROLLBACK;");
                // Throw SQL Error
                throw new exceptions_1.SQLException(error.message || error);
            }
            // Connection Manager has been initiated successfully
            this.Ready = true;
            // Pass database connection to the Locked Schemas
            this.Schemas = this.Schemas.map((schema) => schema.setConnection(this.Id, this));
            return this;
        };
        this.Transaction = async (callback) => {
            try {
                // Start Transaction
                await this.query("BEGIN;");
                // Get Callback Results
                const Results = await callback();
                // Commit Transaction
                await this.query("COMMIT;");
                return Results;
            }
            catch (error) {
                // Rollback Transaction
                await this.query("ROLLBACK;");
                // Throw SQL Error
                throw new exceptions_1.SQLException(error.message || error);
            }
        };
    }
}
exports.ConnectionManager = ConnectionManager;
