"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Connector = void 0;
const promise_1 = require("mysql2/promise");
const exceptions_1 = require("./exceptions");
class Connector {
    constructor(ConnectionOptions, ConnectorOptions = {}) {
        this.ConnectionOptions = ConnectionOptions;
        this.ConnectorOptions = ConnectorOptions;
        // State of connection. Ready will be True if a connection has been initiated.
        this.Ready = false;
        this.connect = async () => {
            try {
                // Check if already initiated a connection.
                if (this.Ready)
                    throw new exceptions_1.SQLConnectionException(`A connection has already been initiated!`);
                // Enable Multiple Statements
                this.ConnectionOptions.multipleStatements = true;
                // Creact SQL Connection
                this.Connection = await promise_1.createConnection(this.ConnectionOptions);
                // Connection has been initiated successfully.
                this.Ready = true;
                return this;
            }
            catch (error) {
                throw new exceptions_1.SQLConnectionException(error);
            }
        };
        this.reConnect = () => this.end().connect();
        this.end = () => {
            // End previous connection
            this.Connection.end();
            // Update Ready State
            this.Ready = false;
            return this;
        };
        this.getConnection = () => {
            if (!this.Ready)
                throw new exceptions_1.SQLConnectionException(`No connection has been initiated yet!`);
            else
                return this.Connection;
        };
        this.getConnectionOptions = () => this.ConnectionOptions;
        this.getConnectorOptions = () => this.ConnectorOptions;
        // Resolve Connector Options
        this.ConnectorOptions.logs = !!this.ConnectorOptions.logs;
        this.ConnectorOptions.sync = !!this.ConnectorOptions.sync;
    }
}
exports.Connector = Connector;
