"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SQLBoolean = void 0;
const base_1 = require("./base");
const exceptions_1 = require("../exceptions");
class SQLBoolean extends base_1.SQLBaseType {
    constructor(options) {
        super("VARCHAR", options || {});
        // Data type ID. Used to defer from other datatypes.
        this._id = "Boolean";
        // Validate input value.
        this.validator = (value) => {
            if (typeof value === "function")
                value = value();
            if (typeof value === "boolean" || value === null)
                return value;
            else
                throw new exceptions_1.SQLValidationException(`Invalid value of type '${typeof value}' for column '${this.columnId}'! Boolean type was expected.`);
        };
        // Returning value resolver
        this.resolver = (value) => {
            if (value === null)
                return value;
            else
                return ((typeof value === "number" && value === 1) ||
                    (typeof value === "string" && value.trim() === "1") ||
                    (typeof value === "string" && value.trim().toLowerCase() === "true") ||
                    (typeof value === "boolean" && value === true));
        };
        if (options && typeof options.validator === "function")
            this.validator = options.validator;
        if (options && typeof options.resolver === "function")
            this.resolver = options.resolver;
    }
}
exports.SQLBoolean = SQLBoolean;
