"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SQLEnum = void 0;
const base_1 = require("./base");
const exceptions_1 = require("../exceptions");
class SQLEnum extends base_1.SQLBaseType {
    constructor(choices, options) {
        super("LONGTEXT", options || {});
        this.choices = choices;
        // Data type ID. Used to defer from other datatypes.
        this._id = "Enum";
        // Validate input value.
        this.validator = (value) => {
            if (typeof value === "function")
                value = value();
            if (this.choices.includes(value) || value === null)
                return value;
            else
                throw new exceptions_1.SQLValidationException(`Invalid value '${value}' for column '${this.columnId}'! Value not found in the Enum choice list. Allowed values are (${this.choices}).`);
        };
        // Returning value resolver
        this.resolver = (value) => {
            if (typeof value.toString === "function" && value !== null)
                return value.toString();
            else
                return value;
        };
        if (options && typeof options.validator === "function")
            this.validator = options.validator;
        if (options && typeof options.resolver === "function")
            this.resolver = options.resolver;
    }
}
exports.SQLEnum = SQLEnum;
