"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SQLNumber = void 0;
const base_1 = require("./base");
class SQLNumber extends base_1.SQLBaseType {
    constructor(options) {
        super("VARCHAR", options || {});
        // Data type ID. Used to defer from other datatypes.
        this._id = "Number";
        // Length of content in the column.
        this.isLength = 15;
        // Decimal Length of Number
        this.decimalLength = 0;
        // Validate input value.
        this.validator = (value) => {
            if (typeof value === "function")
                value = value();
            if (isNaN(value) && value !== null)
                throw new Error(`Invalid value of type '${typeof value}' for column '${this.columnId}'! Number type is expected!`);
            return value;
        };
        // Returning value resolver
        this.resolver = (value) => {
            if (value === null)
                return value;
            if (this.decimalLength === 0)
                return parseInt(value);
            else
                return parseFloat(parseFloat(value).toFixed(this.decimalLength));
        };
        if (options &&
            typeof options.decimalLength === "number" &&
            options.decimalLength > 0)
            this.decimalLength = options.decimalLength;
        if (typeof (options === null || options === void 0 ? void 0 : options.isLength) === "number" && options.isLength > 0)
            this.isLength = options.isLength;
        if (options && typeof options.validator === "function")
            this.validator = options.validator;
        if (options && typeof options.resolver === "function")
            this.resolver = options.resolver;
    }
}
exports.SQLNumber = SQLNumber;
