"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SQLObject = void 0;
const base_1 = require("./base");
const exceptions_1 = require("../exceptions");
class SQLObject extends base_1.SQLBaseType {
    constructor(options) {
        super("LONGTEXT", options || {});
        // Data type ID. Used to defer from other datatypes.
        this._id = "Object";
        // Default value of the column.
        this.defaultValue = {};
        // Validate input value.
        this.validator = (value) => {
            if (typeof value === "function")
                value = value();
            if (typeof value === "object" || value === null)
                return value;
            else
                throw new exceptions_1.SQLValidationException(`Value should be an object type for column '${this.columnId}'!`);
        };
        // Returning value resolver
        this.resolver = (value) => {
            if (typeof value === "string")
                try {
                    return JSON.parse(value);
                }
                catch (error) {
                    return value;
                }
            else
                return value;
        };
        if (options && typeof options.validator === "function")
            this.validator = options.validator;
        if (options && typeof options.resolver === "function")
            this.resolver = options.resolver;
    }
}
exports.SQLObject = SQLObject;
