"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SQLString = void 0;
const base_1 = require("./base");
const exceptions_1 = require("../exceptions");
class SQLString extends base_1.SQLBaseType {
    constructor(options) {
        super("VARCHAR", options || {});
        // Data type ID. Used to defer from other datatypes.
        this._id = "String";
        // Validate input value.
        this.validator = (value) => {
            if (typeof value === "function")
                value = value();
            if (typeof value === "string" ||
                typeof value === "number" ||
                value === null)
                return value;
            else
                throw new exceptions_1.SQLValidationException(`Value should be a string type for column '${this.columnId}'!`);
        };
        // Returning value resolver
        this.resolver = (value) => {
            if (value === null)
                return value;
            if (typeof value.toString === "function")
                return value.toString();
            else
                return value;
        };
        if (options && typeof options.validator === "function")
            this.validator = options.validator;
        if (options && typeof options.resolver === "function")
            this.resolver = options.resolver;
    }
}
exports.SQLString = SQLString;
