import { ConditionsBuilder, ConditionWhere, MultipleCondition, SelectionColumns, SQL_QUERY, StraightCondition } from "./conditions-builder";
import { ConnectionManager } from "./connection-manager";
import { SQLManyRelation, SQLOneRelation } from "./data-types/";
import { SchemaReturnDataType, SchemaRequiredDataType, SchemaType, SchemaDataType } from "./data-types";
import { Schema } from "./schema";
import { SQLFunction, SQLMatch, SQLNotSelection, SQLSelection } from "./utils";
export declare type FilterFlags<Base, Condition> = {
    [Key in keyof Base]: Base[Key] extends Condition ? Key : never;
};
export declare type AllowedKeys<Base, Condition> = FilterFlags<Base, Condition>[keyof Base];
export declare type RenderedSchema<T extends Schema<any>> = {
    name: string;
    schema: T;
};
export declare type ColumnSelection<S extends LockedSchema<any>> = Array<SelectionColumns<S> | "_id" | SQLFunction> | Array<"*"> | SQLNotSelection<SelectionColumns<S> | "_id"> | SQLSelection<SelectionColumns<S> | "_id">;
export declare type SQL_HEADERS = {
    fieldCount: number;
    affectedRows: number;
    insertId: number;
    info: string;
    serverStatus: number;
    warningStatus: number;
    changedRows?: number;
};
export declare class LockedSchema<S extends Schema<any>> {
    Id: string;
    protected Schema: S;
    protected Resetable: boolean;
    protected RecursionDepth: number;
    protected SelectResultsResolverEnabled: boolean;
    protected processMeta: boolean;
    protected ConnectionManager?: ConnectionManager<any>;
    protected ConditionsBuilder: ConditionsBuilder<S>;
    protected SearchMatch: SQLMatch<S>;
    protected FilterMatch: SQLMatch<S>;
    protected Type: string;
    protected SelectionColumns?: ColumnSelection<LockedSchema<S>>;
    protected SelectionConditions?: ConditionWhere<LockedSchema<S>>;
    protected Selection?: {
        custom?: string | ((schemaName: string, query: SQL_QUERY) => string | SQL_QUERY);
        columns?: ColumnSelection<LockedSchema<S>>;
        conditions?: ConditionWhere<LockedSchema<S>>;
    };
    protected Relations: {
        [key: string]: LockedSchema<any>;
    };
    protected Resources: {};
    protected forEach: <O extends Record<any, any>, T>(Object: O, callback: (value: O[keyof O], key: keyof O, Object: O) => T) => any[];
    protected ResolveSelectResults: (Results: Array<SchemaReturnDataType<SchemaType<S>>>) => Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
    constructor(Id: string, Schema: S);
    getBaseSchema: () => S;
    setConnection: (ConnectionManager: ConnectionManager<any>) => this;
    hasConnection: () => boolean;
    query: (sql: string, values?: any[] | undefined) => Promise<[import("mysql2/typings/mysql/lib/protocol/packets/OkPacket") | import("mysql2/typings/mysql/lib/protocol/packets/ResultSetHeader") | import("mysql2/typings/mysql/lib/protocol/packets/RowDataPacket")[] | import("mysql2/typings/mysql/lib/protocol/packets/RowDataPacket")[][] | import("mysql2/typings/mysql/lib/protocol/packets/OkPacket")[], import("mysql2/typings/mysql/lib/protocol/packets/FieldPacket")[]]> | undefined;
    setType: (Type: string) => this;
    getType: () => string;
    getSchemaName: () => string;
    withRelation: {
        (Relation: AllowedKeys<SchemaType<S>, SQLOneRelation<any> | SQLManyRelation<any>> | Array<AllowedKeys<SchemaType<S>, SQLOneRelation<any> | SQLManyRelation<any>>>, Schema?: LockedSchema<any>): LockedSchema<S>;
        (Relation: {
            [key in AllowedKeys<SchemaType<S>, SQLOneRelation<any> | SQLManyRelation<any>>]?: LockedSchema<any>;
        }): LockedSchema<S>;
    };
    resetRelation: () => this;
    setResources: (resources: any) => this;
    resetResources: () => this;
    enableProcessMeta: () => this;
    disableProcessMeta: () => this;
    where: {
        (WHERE: SQLMatch<S>): LockedSchema<S>;
        (WHERE: MultipleCondition<LockedSchema<S>>): LockedSchema<S>;
        (WHERE: StraightCondition<LockedSchema<S>>): LockedSchema<S>;
    };
    having: {
        (HAVING: SQLMatch<S>): LockedSchema<S>;
        (HAVING: MultipleCondition<LockedSchema<S>>): LockedSchema<S>;
        (HAVING: StraightCondition<LockedSchema<S>>): LockedSchema<S>;
    };
    orderBy: (COLUMN: SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction, SORT?: "ASC" | "DESC" | undefined) => this;
    sort: (SORT: "ASC" | "DESC") => this;
    asc: (COLUMN?: SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction) => this;
    desc: (COLUMN?: SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction) => this;
    groupBy: (GROUP: Array<SelectionColumns<LockedSchema<S>> | "_id" | SQLFunction>) => this;
    offset: (OFFSET: number) => this;
    limit: (LIMIT: number) => this;
    columns: (Columns?: ColumnSelection<LockedSchema<S>> | undefined) => this;
    resetColumns: () => this;
    conditions: (Conditions?: ConditionWhere<LockedSchema<S>> | undefined) => this;
    resetConditions: () => this;
    selection: {
        (Selection?: ColumnSelection<LockedSchema<S>>): LockedSchema<S>;
        (Selection?: {
            custom?: string | ((schemaName: string, query: SQL_QUERY) => string | SQL_QUERY);
            columns?: ColumnSelection<LockedSchema<S>>;
            conditions?: ConditionWhere<LockedSchema<S>>;
        }): LockedSchema<S>;
    };
    resetSelection: () => this;
    searchMode: (mode?: "boolean" | "naturalLanguage" | undefined) => this;
    searchWith: (relevance?: "queryExpansion" | undefined) => this;
    searchWhere: (conditions: ConditionWhere<LockedSchema<S>>, operation?: "OR" | "AND" | undefined) => this;
    search: (search: string) => this;
    filterMode: (mode?: "boolean" | "naturalLanguage" | undefined) => this;
    filterWith: (relevance?: "queryExpansion" | undefined) => this;
    filterWhere: (conditions: ConditionWhere<LockedSchema<S>>, operation?: "OR" | "AND" | undefined) => this;
    filter: (search: string) => this;
    select: {
        (Selection?: ColumnSelection<LockedSchema<S>>, throwError?: boolean): Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
        (Selection?: {
            custom?: string | ((schemaName: string, query: SQL_QUERY) => string | SQL_QUERY);
            columns?: ColumnSelection<LockedSchema<S>>;
            conditions?: ConditionWhere<LockedSchema<S>>;
        }, throwError?: boolean): Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
    };
    selectOrFail: {
        (Selection?: ColumnSelection<LockedSchema<S>>): Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
        (Selection?: {
            custom?: string | ((schemaName: string, query: SQL_QUERY) => string | SQL_QUERY);
            columns?: ColumnSelection<LockedSchema<S>>;
            conditions?: ConditionWhere<LockedSchema<S>>;
        }): Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
    };
    selectOne: {
        (Selection?: ColumnSelection<LockedSchema<S>>, throwError?: boolean): Promise<SchemaReturnDataType<SchemaType<S>>>;
        (Selection?: {
            custom?: string | ((schemaName: string, query: SQL_QUERY) => string | SQL_QUERY);
            columns?: ColumnSelection<LockedSchema<S>>;
            conditions?: ConditionWhere<LockedSchema<S>>;
        }, throwError?: boolean): Promise<SchemaReturnDataType<SchemaType<S>>>;
    };
    selectOneOrFail: {
        (Selection?: ColumnSelection<LockedSchema<S>>): Promise<SchemaReturnDataType<SchemaType<S>>>;
        (Selection?: {
            custom?: string | ((schemaName: string, query: SQL_QUERY) => string | SQL_QUERY);
            columns?: ColumnSelection<LockedSchema<S>>;
            conditions?: ConditionWhere<LockedSchema<S>>;
        }): Promise<SchemaReturnDataType<SchemaType<S>>>;
    };
    count: () => Promise<number>;
    sum: (column: SelectionColumns<LockedSchema<S>>) => Promise<number>;
    avg: (column: SelectionColumns<LockedSchema<S>>) => Promise<number>;
    insert: {
        (data: SchemaRequiredDataType<SchemaType<S>>, headers?: boolean): Promise<SchemaReturnDataType<SchemaType<S>>>;
        (data: Array<SchemaRequiredDataType<SchemaType<S>>>, headers?: boolean): Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
    };
    insertIfAble: {
        (data: SchemaRequiredDataType<SchemaType<S>>, headers?: boolean): Promise<SchemaReturnDataType<SchemaType<S>>> | undefined;
        (data: Array<SchemaRequiredDataType<SchemaType<S>>>, headers?: boolean): Promise<Array<SchemaReturnDataType<SchemaType<S>>>> | undefined;
    };
    updateOrInsert: (data: SchemaDataType<SchemaType<S>>, rest: SchemaDataType<SchemaType<S>>, throwError?: boolean) => Promise<SchemaReturnDataType<SchemaType<S>>>;
    update: (data: SchemaDataType<SchemaType<S>>, throwError?: boolean) => Promise<SQL_HEADERS>;
    updateOrFail: (data: SchemaDataType<SchemaType<S>>) => Promise<SQL_HEADERS>;
    selectAndUpdate: (data: SchemaDataType<SchemaType<S>>, throwError?: boolean) => Promise<SchemaReturnDataType<SchemaType<S>>[]>;
    selectOneAndUpdate: (data: SchemaDataType<SchemaType<S>>, throwError?: boolean) => Promise<SchemaReturnDataType<SchemaType<S>>>;
    updateAndSelect: (data: SchemaDataType<SchemaType<S>>, throwError?: boolean) => Promise<Array<SchemaReturnDataType<SchemaType<S>>>>;
    updateAndSelectOne: (data: SchemaDataType<SchemaType<S>>, throwError?: boolean) => Promise<SchemaReturnDataType<SchemaType<S>>>;
    delete: (throwError?: boolean) => Promise<SQL_HEADERS>;
    deleteOrFail: () => Promise<SQL_HEADERS>;
    selectAndDelete: (throwError?: boolean) => Promise<SchemaReturnDataType<SchemaType<S>>[]>;
    selectOneAndDelete: (throwError?: boolean) => Promise<SchemaReturnDataType<SchemaType<S>>>;
    empty: () => Promise<SQL_HEADERS>;
    isResetable: () => this;
    notResetable: () => this;
    reset: () => this;
}
