"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LockedSchema = void 0;
const conditions_builder_1 = require("./conditions-builder");
const connection_manager_1 = require("./connection-manager");
const data_types_1 = require("./data-types/");
const exceptions_1 = require("./exceptions");
const utils_1 = require("./utils");
class LockedSchema {
    constructor(Id, Schema) {
        this.Id = Id;
        this.Schema = Schema;
        this.Resetable = true;
        this.RecursionDepth = 0;
        this.SelectResultsResolverEnabled = true;
        this.processMeta = true;
        this.ConditionsBuilder = new conditions_builder_1.ConditionsBuilder();
        this.SearchMatch = new utils_1.SQLMatch(this.Schema.SearchableColumns);
        this.FilterMatch = new utils_1.SQLMatch(this.Schema.SearchableColumns);
        this.Relations = {};
        this.Resources = {};
        this.forEach = (Object, callback) => {
            const Results = [];
            for (let key in Object) {
                if (Object.hasOwnProperty(key)) {
                    const Result = callback(Object[key], key, Object);
                    if (typeof Result === "boolean") {
                        if (!Result)
                            break;
                        else
                            continue;
                    }
                    else
                        Results.push(Result);
                }
            }
            return Results;
        };
        this.ResolveSelectResults = async (Results) => {
            if (!this.SelectResultsResolverEnabled)
                return Results;
            let ResolvedResults = await Promise.all(Results.map(async (Row) => {
                Object.keys(Row).map((column) => {
                    try {
                        if (column !== "_id" && column !== "_headers") {
                            const ColumnType = this.Schema.Data[column];
                            // Sanitize Result
                            if (ColumnType instanceof data_types_1.SQLString ||
                                ColumnType instanceof data_types_1.SQLContent ||
                                ColumnType instanceof data_types_1.SQLEnum ||
                                ColumnType instanceof data_types_1.SQLBoolean ||
                                ColumnType instanceof data_types_1.SQLNumber ||
                                ColumnType instanceof data_types_1.SQLObject)
                                Row[column] = ColumnType.resolver(Row[column]);
                        }
                    }
                    catch (error) {
                        // Do nothing...
                    }
                });
                // Resolve Relations
                const Relations = Object.keys(this.Relations);
                if (Relations.length) {
                    await Promise.all(Relations.map(async (Relation) => {
                        const RelationColumn = this.Schema.Data[Relation];
                        if (RelationColumn instanceof data_types_1.SQLOneRelation ||
                            RelationColumn instanceof data_types_1.SQLManyRelation) {
                            if (typeof RelationColumn.Column === "string") {
                                if (Row[typeof RelationColumn.ForeignColumn === "string"
                                    ? RelationColumn.ForeignColumn
                                    : RelationColumn.Column] !== undefined) {
                                    if (this.Relations[Relation] instanceof LockedSchema) {
                                        // @ts-ignore
                                        Row[Relation] = await this.Relations[Relation][RelationColumn instanceof data_types_1.SQLOneRelation
                                            ? "selectOne"
                                            : "select"]({
                                            conditions: {
                                                [RelationColumn.Column]: Row[typeof RelationColumn.ForeignColumn === "string"
                                                    ? RelationColumn.ForeignColumn
                                                    : RelationColumn.Column],
                                            },
                                        });
                                        // Resolve Results
                                        if (typeof RelationColumn.resolver === "function")
                                            // @ts-ignore
                                            Row[Relation] = RelationColumn.resolver(Row[Relation], this.Relations[Relation], this.Resources);
                                    }
                                    else
                                        throw new exceptions_1.SQLException(`No schema has been provided on the relation!`);
                                }
                            }
                            else if (RelationColumn.Column instanceof Array) {
                                if (RelationColumn.Column.filter((column, index) => typeof Row[RelationColumn.ForeignColumn instanceof Array
                                    ? RelationColumn.ForeignColumn[index]
                                    : column] !== "undefined").length === RelationColumn.Column.length) {
                                    // @ts-ignore
                                    Row[Relation] = await this.Relations[Relation][RelationColumn instanceof data_types_1.SQLOneRelation
                                        ? "selectOne"
                                        : "select"]({
                                        conditions: RelationColumn.Column.reduce((p, c, index) => (Object.assign(Object.assign({}, p), {
                                            [c]: Row[RelationColumn.ForeignColumn instanceof Array
                                                ? RelationColumn.ForeignColumn[index]
                                                : c],
                                        })), {}),
                                    });
                                    // Resolve Results
                                    if (typeof RelationColumn.resolver === "function")
                                        // @ts-ignore
                                        Row[Relation] = RelationColumn.resolver(Row[Relation], this.Relations[Relation], this.Resources);
                                }
                            }
                        }
                        else
                            throw new exceptions_1.SQLException(`Invalid column type for relation!`);
                    }));
                }
                return Row;
            }));
            // Resolve Meta Data
            if (this.processMeta)
                ResolvedResults = await Promise.all(ResolvedResults.map(async (Row) => {
                    var _a;
                    const MetaColumns = this.Schema.metaResolvers;
                    for (let ColumnKey in MetaColumns)
                        if (typeof MetaColumns[ColumnKey].resolver === "function")
                            if (!(MetaColumns[ColumnKey].dependencies instanceof Array) ||
                                ((_a = MetaColumns[ColumnKey].dependencies) === null || _a === void 0 ? void 0 : _a.reduce((p, c) => {
                                    if (Row[c] === undefined)
                                        return false;
                                    else
                                        return p;
                                }, true)))
                                // @ts-ignore
                                Row[ColumnKey] = await MetaColumns[ColumnKey].resolver(Row, this, this.Resources);
                    return Row;
                }));
            // Reset Schema
            this.reset();
            return ResolvedResults;
        };
        this.getBaseSchema = () => this.Schema;
        this.setConnection = (ConnectionManager) => {
            this.ConnectionManager = ConnectionManager;
            return this;
        };
        this.hasConnection = () => this.ConnectionManager instanceof connection_manager_1.ConnectionManager;
        this.query = (sql, values) => {
            var _a;
            if (!this.hasConnection())
                throw new exceptions_1.SQLException(`No connection has been found on '${this.Schema.Name}' Schema!`);
            else
                try {
                    return (_a = this.ConnectionManager) === null || _a === void 0 ? void 0 : _a.query(sql, values);
                }
                catch (error) {
                    throw new exceptions_1.SQLException(error);
                }
        };
        this.setType = (Type) => {
            if (!this.Schema.Types.length)
                throw new exceptions_1.SQLException(`No types are available on the Schema!`);
            else if (!this.Schema.Types.includes(Type))
                throw new exceptions_1.SQLException(`Invalid schema type name '${Type}' has been provided! Type not found.`);
            else
                this.Type = Type;
            return this;
        };
        this.getType = () => this.Type;
        this.getSchemaName = () => [this.Schema.Prefix, this.Schema.Name, this.Type]
            .filter((v) => v)
            .join("-");
        this.withRelation = (Relation, Schema) => {
            if (Relation instanceof Array)
                Relation.forEach((relation) => this.withRelation(relation));
            else if (typeof Relation === "string") {
                if (!this.Schema.hasColumn(Relation))
                    throw new exceptions_1.SQLException(`Invalid relation '${Relation}' has been provided!`);
                else if (Schema && !(Schema instanceof LockedSchema))
                    throw new exceptions_1.SQLException(`Invalid schema has been provided for '${Relation}' relation!`);
                else if (!(this.Schema.Data[Relation].Schema(this.Id) instanceof LockedSchema))
                    throw new exceptions_1.SQLException(`Invalid relation '${Relation}' has been provided! No or Invalid relation found!`);
                else
                    this.Relations[Relation] =
                        Schema ||
                            this.Schema.Data[Relation].Schema(this.Id);
            }
            else if (typeof Relation === "object")
                this.forEach(Relation, (schema, relation) => this.withRelation(relation, schema));
            return this;
        };
        this.resetRelation = () => {
            this.Relations = {};
            return this;
        };
        this.setResources = (resources) => {
            this.Resources = resources;
            return this;
        };
        this.resetResources = () => {
            this.Resources = {};
            return this;
        };
        this.enableProcessMeta = () => {
            this.processMeta = true;
            return this;
        };
        this.disableProcessMeta = () => {
            this.processMeta = false;
            return this;
        };
        this.where = (WHERE) => {
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { WHERE }));
            return this;
        };
        this.having = (HAVING) => {
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { HAVING }));
            return this;
        };
        this.orderBy = (COLUMN, SORT) => {
            var _a;
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { ORDER: {
                    COLUMN,
                    SORT: SORT || ((_a = this.ConditionsBuilder.ORDER) === null || _a === void 0 ? void 0 : _a.SORT),
                } }));
            return this;
        };
        this.sort = (SORT) => {
            var _a;
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { ORDER: { COLUMN: (_a = this.ConditionsBuilder.ORDER) === null || _a === void 0 ? void 0 : _a.COLUMN, SORT } }));
            return this;
        };
        this.asc = (COLUMN = "_id") => this.orderBy(COLUMN, "ASC");
        this.desc = (COLUMN = "_id") => this.orderBy(COLUMN, "DESC");
        this.groupBy = (GROUP) => {
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { GROUP }));
            return this;
        };
        this.offset = (OFFSET) => {
            var _a;
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { LIMIT: { OFFSET, NUMBER: (_a = this.ConditionsBuilder.LIMIT) === null || _a === void 0 ? void 0 : _a.NUMBER } }));
            return this;
        };
        this.limit = (LIMIT) => {
            var _a;
            this.ConditionsBuilder.conditions(Object.assign(Object.assign({}, this.ConditionsBuilder), { LIMIT: {
                    OFFSET: (_a = this.ConditionsBuilder.LIMIT) === null || _a === void 0 ? void 0 : _a.OFFSET,
                    NUMBER: LIMIT,
                } }));
            return this;
        };
        this.columns = (Columns) => {
            this.SelectionColumns = Columns;
            return this;
        };
        this.resetColumns = () => {
            delete this.SelectionColumns;
            return this;
        };
        this.conditions = (Conditions) => {
            this.SelectionConditions = Conditions;
            return this;
        };
        this.resetConditions = () => {
            delete this.SelectionConditions;
            return this;
        };
        this.selection = (Selection) => {
            if (typeof Selection !== "object")
                throw new exceptions_1.SQLException(`Invalid selection has been provided!`);
            this.Selection =
                Selection instanceof Array ? { columns: Selection } : Selection;
            return this;
        };
        this.resetSelection = () => {
            delete this.Selection;
            return this;
        };
        this.searchMode = (...params) => {
            this.SearchMatch.mode(...params);
            return this;
        };
        this.searchWith = (...params) => {
            this.SearchMatch.with(...params);
            return this;
        };
        this.searchWhere = (...params) => {
            this.SearchMatch.where(...params);
            return this;
        };
        this.search = (search) => {
            this.SearchMatch.search = search;
            this.where(this.SearchMatch);
            return this;
        };
        this.filterMode = (...params) => {
            this.FilterMatch.mode(...params);
            return this;
        };
        this.filterWith = (...params) => {
            this.FilterMatch.with(...params);
            return this;
        };
        this.filterWhere = (...params) => {
            this.FilterMatch.where(...params);
            return this;
        };
        this.filter = (search) => {
            this.FilterMatch.search = search;
            this.conditions(this.FilterMatch);
            return this;
        };
        this.select = async (Selection = [], throwError = false) => {
            try {
                let QueryMethod = `SELECT *`;
                let QueryConditions = this.ConditionsBuilder.render();
                // Get Default Selection
                if (typeof this.Selection === "object") {
                    if ((Selection instanceof Array && Selection.length) ||
                        (typeof Selection === "object" && !(Selection instanceof Array))) {
                        console.log("WARNING! ::::::::::::");
                        console.log(`Your inline provided selection on select operation at schema '${this.Schema.Name}' has been overridden! This might cause invalid results. Please review your query & results.`, "\n");
                    }
                    // Override Inline Selection
                    Selection = this.Selection;
                }
                else if (!Selection)
                    Selection = [];
                // Override Selection Columns
                if (this.SelectionColumns instanceof Array ||
                    this.SelectionColumns instanceof utils_1.SQLSelection ||
                    this.SelectionColumns instanceof utils_1.SQLNotSelection) {
                    if (Selection instanceof Array)
                        Selection = this.SelectionColumns;
                    else if (typeof Selection === "object")
                        Selection.columns = this.SelectionColumns;
                }
                // Override Selection Conditions
                if (typeof this.SelectionConditions === "object") {
                    if (Selection instanceof Array) {
                        Selection = {
                            columns: Selection,
                            conditions: this.SelectionConditions,
                        };
                    }
                    else if (typeof Selection === "object")
                        Selection.conditions = this.SelectionConditions;
                }
                if (Selection instanceof utils_1.SQLNotSelection) {
                    // Validate Selection Columns
                    const Columns = Selection.get().map((column) => {
                        if (column instanceof utils_1.SQLFunction)
                            throw new exceptions_1.SQLException("SQL Function is not allowed for Not Selection!");
                        else if (!this.Schema.hasColumn(column))
                            throw new exceptions_1.SQLException(`Invalid column name '${column}' in selection list!`);
                        else
                            return column;
                    });
                    const WhiteListColumns = this.Schema.columns()
                        .filter((column) => !Columns.includes(column))
                        .filter((column) => !this.Schema.Data[column].isVirtual);
                    // Create Query
                    QueryMethod = `SELECT ${WhiteListColumns.length ? `\`${WhiteListColumns.join("`, `")}\`` : "*"} FROM \`${this.getSchemaName()}\``;
                }
                else if (Selection instanceof Array ||
                    Selection instanceof utils_1.SQLSelection) {
                    if (Selection instanceof Array && Selection.length) {
                        // Validate Selection Columns
                        const Columns = Selection.map((column) => {
                            if (column instanceof utils_1.SQLFunction)
                                return column.toString();
                            else if (column === "*")
                                return "*";
                            else if (!this.Schema.hasColumn(column))
                                throw new exceptions_1.SQLException(`Invalid column name '${column}' in selection list!`);
                            else
                                return `\`${column}\``;
                        });
                        // Create Query
                        QueryMethod = `SELECT ${Columns.length ? Columns : "*"} FROM \`${this.getSchemaName()}\``;
                    }
                    else
                        return this.resetSelection().select(this.Schema.columns()
                            .filter((column) => this.Schema.Data[column].isPublic ||
                            (Selection instanceof utils_1.SQLSelection &&
                                Selection.get().includes(column)))
                            .filter((column) => !this.Schema.Data[column].isVirtual), throwError);
                }
                else if (typeof Selection === "object" &&
                    !["string", "function"].includes(typeof Selection.custom)) {
                    // Resolve Sub Condition
                    if (typeof Selection.conditions !== "object")
                        return this.resetSelection().select(Selection.columns, throwError);
                    // Create Nested Condition
                    const TempQuery = new conditions_builder_1.ConditionsBuilder()
                        .conditions({ WHERE: Selection.conditions })
                        .render();
                    QueryConditions.values = [
                        ...TempQuery.values,
                        ...QueryConditions.values,
                    ];
                    // Validate Selection Columns
                    const SelectionColumns = Selection.columns;
                    const Columns = SelectionColumns instanceof utils_1.SQLNotSelection
                        ? SelectionColumns.get().map((column) => {
                            if (column instanceof utils_1.SQLFunction)
                                throw new exceptions_1.SQLException("SQL Function is not allowed for Not Selection!");
                            else if (!this.Schema.hasColumn(column))
                                throw new exceptions_1.SQLException(`Invalid column name '${column}' in selection list!`);
                            else
                                return column;
                        })
                        : SelectionColumns instanceof Array
                            ? SelectionColumns.map((column) => {
                                if (column instanceof utils_1.SQLFunction)
                                    return column.toString();
                                else if (column === "*")
                                    return "*";
                                else if (!this.Schema.hasColumn(column))
                                    throw new exceptions_1.SQLException(`Invalid column name '${column}' in selection list!`);
                                else
                                    return `\`${column}\``;
                            })
                            : SelectionColumns instanceof utils_1.SQLSelection
                                ? this.Schema.columns()
                                    .filter((column) => this.Schema.Data[column].isPublic ||
                                    (SelectionColumns instanceof utils_1.SQLSelection &&
                                        SelectionColumns.get().includes(column)))
                                    .filter((column) => !this.Schema.Data[column].isVirtual)
                                : [];
                    // Create Nested Query
                    QueryMethod = `SELECT ${Columns.length
                        ? Columns
                        : `\`${this.Schema.columns()
                            .filter((column) => this.Schema.Data[column].isPublic)
                            .filter((column) => !this.Schema.Data[column].isVirtual)
                            .join("`, `")}\``} FROM (SELECT * FROM \`${this.getSchemaName()}\` ${TempQuery.sql}) AS \`TEMP_SELECTION_1\``;
                }
                else if (["string", "function"].includes(typeof Selection.custom)) {
                    if (typeof Selection.custom === "string")
                        QueryMethod = `SELECT * FROM (${Selection.custom.replace(/(;\s*)$/, "")}) AS TEMP_SELECTION_1`;
                    else if (typeof Selection.custom === "function") {
                        // Execute Selection Query
                        const SelectionQuery = Selection.custom(this.getSchemaName(), QueryConditions);
                        // Check Selection Query
                        if (typeof SelectionQuery === "string") {
                            QueryMethod = SelectionQuery;
                            QueryConditions = { sql: "", values: [] };
                        }
                        else if (typeof SelectionQuery === "object") {
                            QueryMethod = "";
                            QueryConditions = SelectionQuery;
                        }
                        else
                            throw new exceptions_1.SQLException(`Invalid query has been provided on the Select method!`);
                    }
                }
                // Prepare Query
                const Results = await this.query(`${QueryMethod + (QueryConditions.sql ? " " + QueryConditions.sql : "")};`, QueryConditions.values || []);
                // Return Result
                if (Results && Results instanceof Array && Results[0] instanceof Array) {
                    if (Results[0].length)
                        return await this.ResolveSelectResults(Results[0]);
                    else if (throwError)
                        throw new exceptions_1.SQLException(`No results found!`);
                    else
                        return [];
                }
                else
                    throw new exceptions_1.SQLException("Unexpected result pattern!");
            }
            catch (error) {
                this.reset();
                throw error;
            }
        };
        this.selectOrFail = (Selection) => this.select(Selection, true);
        this.selectOne = async (Selection, throwError = false) => (await this.limit(1).select(Selection, throwError))[0];
        this.selectOneOrFail = (Selection) => this.selectOne(Selection, true);
        this.count = async () => {
            // Reset Relations
            this.resetRelation();
            // Reset Default Selection
            this.resetSelection();
            // Reset Default Columns
            this.resetColumns();
            // Disable Select Results Resolver To Avoid Any Errors
            this.SelectResultsResolverEnabled = false;
            // Count Rows
            const Count = (await this.selectOne([new utils_1.SQLFunction("COUNT(*) as `count`")])).count;
            // Enable Select Results Resolver Back
            this.SelectResultsResolverEnabled = true;
            return Count;
        };
        this.sum = async (column) => {
            // Reset Relations
            this.resetRelation();
            // Reset Default Selection
            this.resetSelection();
            // Reset Default Columns
            this.resetColumns();
            // Disable Select Results Resolver To Avoid Any Errors
            this.SelectResultsResolverEnabled = false;
            // Get Sum Rows
            const Sum = (await this.selectOne([new utils_1.SQLFunction(`SUM(\`${column}\`) as \`sum\``)])).sum;
            // Enable Select Results Resolver Back
            this.SelectResultsResolverEnabled = true;
            return Sum;
        };
        this.avg = async (column) => {
            // Reset Relations
            this.resetRelation();
            // Reset Default Selection
            this.resetSelection();
            // Reset Default Columns
            this.resetColumns();
            // Disable Select Results Resolver To Avoid Any Errors
            this.SelectResultsResolverEnabled = false;
            // Get Avg Rows
            const Avg = (await this.selectOne([new utils_1.SQLFunction(`AVG(\`${column}\`) as \`avg\``)])).avg;
            // Enable Select Results Resolver Back
            this.SelectResultsResolverEnabled = true;
            return Avg;
        };
        // @ts-ignore
        this.insert = async (data, headers = false) => {
            var _a;
            try {
                // @ts-ignore
                if (data instanceof Array)
                    return await Promise.all(data.map((_) => this.insert(_, headers)));
                else if (typeof data === "object") {
                    // Increment Recursion Depth
                    this.RecursionDepth++;
                    // Validated Data
                    const Data = {};
                    const Relations = {};
                    // Validate Data Columns
                    await Promise.all(this.forEach(this.Schema.Data, async (columnType, columnName) => {
                        if (columnType instanceof data_types_1.SQLMeta) {
                            // Do nothing...
                        }
                        else if (columnType instanceof data_types_1.SQLOneRelation ||
                            columnType instanceof data_types_1.SQLManyRelation) {
                            if (typeof data[columnName] === "object")
                                Relations[columnName] = {
                                    column: columnType,
                                    data: data[columnName],
                                };
                        }
                        else if (typeof data[columnName] !== "undefined") {
                            if (typeof columnType.validator === "function") {
                                try {
                                    Data[columnName] = await columnType.validator(data[columnName]);
                                }
                                catch (error) {
                                    throw new exceptions_1.SQLValidationException(error.message || error);
                                }
                            }
                            else
                                Data[columnName] = data[columnName];
                        }
                        else if (typeof columnType.defaultValue !== "undefined") {
                            if (columnType.defaultValue instanceof utils_1.SQLFunction)
                                Data[columnName] = columnType.defaultValue;
                            else if (typeof columnType.defaultValue === "function") {
                                const Value = await columnType.defaultValue();
                                if (typeof Value !== "undefined")
                                    Data[columnName] = Value;
                            }
                            else
                                Data[columnName] = columnType.defaultValue;
                        }
                    }));
                    // Prepare Query String
                    const DataKeys = Object.keys(Data);
                    const DataValues = [];
                    if (!DataKeys.length)
                        throw new exceptions_1.SQLException(`Nothing has been provided to insert!`);
                    const QueryString = `INSERT INTO \`${this.getSchemaName()}\` (${DataKeys.map((column) => `\`${column}\``)}) VALUES (${DataKeys.map((column) => {
                        if (Data[column] instanceof utils_1.SQLFunction)
                            return Data[column].toString();
                        else if (typeof Data[column] === "object" && Data[column] !== null) {
                            DataValues.push(JSON.stringify(Data[column]));
                            return "?";
                        }
                        else {
                            DataValues.push(Data[column]);
                            return "?";
                        }
                    })});`;
                    // Insert Data
                    const Result = (await this.query(QueryString, DataValues));
                    // Resolve Response
                    if (Result &&
                        Result instanceof Array &&
                        Result[0] &&
                        Result[0].insertId) {
                        const Inserted = await this.where({
                            _id: Result[0].insertId,
                        }).selectOneOrFail();
                        // Add Headers If Required
                        if (headers)
                            Inserted._headers = Result[0];
                        // Insert Relational Data
                        await Promise.all(this.forEach(Relations, async (relationObject, column) => {
                            if (typeof relationObject.column.Column === "string") {
                                const Column = relationObject.column.Column;
                                const ForeignColumn = relationObject.column.ForeignColumn;
                                if (typeof Inserted[typeof ForeignColumn === "string" ? ForeignColumn : Column] !== "undefined") {
                                    // Assign Relational Key
                                    if (relationObject.data instanceof Array) {
                                        relationObject.data = relationObject.data.map((row) => {
                                            row[Column] =
                                                Inserted[typeof ForeignColumn === "string"
                                                    ? ForeignColumn
                                                    : Column];
                                            return row;
                                        });
                                    }
                                    else
                                        relationObject.data[Column] =
                                            Inserted[typeof ForeignColumn === "string"
                                                ? ForeignColumn
                                                : Column];
                                    // Insert & Get Data
                                    // @ts-ignore
                                    Inserted[column] = await relationObject.column
                                        .Schema(this.Id)
                                        .insert(relationObject.data, headers);
                                }
                            }
                            else if (relationObject.column.Column instanceof Array) {
                                const Columns = relationObject.column.Column;
                                const ForeignColumn = relationObject.column.ForeignColumn;
                                if (Columns.filter((col, index) => typeof Inserted[ForeignColumn instanceof Array
                                    ? ForeignColumn[index]
                                    : col] !== "undefined").length === Columns.length) {
                                    // Assign Relational Key
                                    if (relationObject.data instanceof Array) {
                                        relationObject.data = relationObject.data.map((row) => {
                                            Columns.forEach((col, index) => (row[col] =
                                                Inserted[ForeignColumn instanceof Array
                                                    ? ForeignColumn[index]
                                                    : col]));
                                            return row;
                                        });
                                    }
                                    else
                                        Columns.forEach((col, index) => (relationObject.data[col] =
                                            Inserted[ForeignColumn instanceof Array
                                                ? ForeignColumn[index]
                                                : col]));
                                    // Insert & Get Data
                                    // @ts-ignore
                                    Inserted[column] = await relationObject.column
                                        .Schema(this.Id)
                                        .insert(relationObject.data, headers);
                                }
                            }
                        }));
                        // Check Recursion Depth
                        this.RecursionDepth--;
                        if (this.RecursionDepth === 0)
                            this.reset();
                        // Resolve Meta Data
                        if (this.processMeta) {
                            const MetaColumns = this.Schema.metaResolvers;
                            for (let ColumnKey in MetaColumns)
                                if (typeof MetaColumns[ColumnKey].resolver === "function")
                                    if (!(MetaColumns[ColumnKey].dependencies instanceof Array) ||
                                        ((_a = MetaColumns[ColumnKey].dependencies) === null || _a === void 0 ? void 0 : _a.reduce((p, c) => {
                                            if (Inserted[c] === undefined)
                                                return false;
                                            else
                                                return p;
                                        }, true)))
                                        // @ts-ignore
                                        Inserted[ColumnKey] = await MetaColumns[ColumnKey].resolver(Inserted, this, this.Resources);
                        }
                        return Inserted;
                    }
                    else
                        throw new exceptions_1.SQLException("Invalid or unknown Insert response!");
                }
                else
                    throw new exceptions_1.SQLException(`Cannot perform insert operation! Invalid data of type '${typeof data}' provided!`);
            }
            catch (error) {
                this.reset();
                throw error;
            }
        };
        // @ts-ignore
        this.insertIfAble = async (data, headers = false) => {
            try {
                return await this.insert(data, headers);
            }
            catch (error) {
                if (error instanceof exceptions_1.SQLValidationException)
                    throw error;
                return;
            }
        };
        this.updateOrInsert = async (data, rest, throwError = false) => {
            if (await this.notResetable().count())
                return this.updateAndSelectOne(data, throwError);
            else
                return this.insert(Object.assign(Object.assign({}, data), rest), throwError);
        };
        this.update = async (data, throwError = false) => {
            try {
                // Get Conditions
                const QueryConditions = this.ConditionsBuilder.render(false);
                // Validated Data
                const Data = {};
                // Validate Data Columns
                await Promise.all(this.forEach(this.Schema.Data, async (columnType, columnName) => {
                    if (columnType instanceof data_types_1.SQLMeta) {
                        // Do nothing...
                    }
                    else if (columnType instanceof data_types_1.SQLOneRelation ||
                        columnType instanceof data_types_1.SQLManyRelation) {
                        // Do nothing...
                    }
                    else if (typeof data[columnName] !== "undefined") {
                        if (typeof columnType.validator === "function") {
                            try {
                                Data[columnName] = await columnType.validator(data[columnName]);
                            }
                            catch (error) {
                                throw new exceptions_1.SQLValidationException(error.message || error);
                            }
                        }
                        else
                            Data[columnName] = data[columnName];
                    }
                    else if (typeof columnType.updatedValue !== "undefined") {
                        if (columnType.updatedValue instanceof utils_1.SQLFunction)
                            Data[columnName] = columnType.updatedValue;
                        else if (typeof columnType.updatedValue === "function") {
                            const Value = await columnType.updatedValue();
                            if (typeof Value !== "undefined")
                                Data[columnName] = Value;
                        }
                        else
                            Data[columnName] = columnType.updatedValue;
                    }
                }));
                // Prepare Query String
                const DataKeys = Object.keys(Data);
                const DataValues = [];
                if (!DataKeys.length)
                    throw new exceptions_1.SQLException(`Nothing has been provided to update!`);
                const QueryString = `UPDATE \`${this.getSchemaName()}\` SET ${DataKeys.map((column) => {
                    if (Data[column] instanceof utils_1.SQLFunction)
                        return Data[column].toString();
                    else if (typeof Data[column] === "object" && Data[column] !== null) {
                        DataValues.push(JSON.stringify(Data[column]));
                        return `\`${column}\` = ?`;
                    }
                    else {
                        DataValues.push(Data[column]);
                        return `\`${column}\` = ?`;
                    }
                }).join(", ")} ${QueryConditions.sql};`;
                // Update Data
                const Result = (await this.query(QueryString, [
                    ...DataValues,
                    ...QueryConditions.values,
                ]));
                // Reset Schema
                this.reset();
                // Resolve Response
                if (Result &&
                    Result instanceof Array &&
                    Result[0] &&
                    typeof Result[0].affectedRows !== "undefined") {
                    if (throwError && !Result[0].affectedRows)
                        throw new exceptions_1.SQLException("Nothing has been updated!");
                    else
                        return Result[0];
                }
                else
                    throw new exceptions_1.SQLException("Invalid or unknown update response!");
            }
            catch (error) {
                this.reset();
                throw error;
            }
        };
        this.updateOrFail = (data) => this.update(data, true);
        this.selectAndUpdate = async (data, throwError = false) => {
            try {
                // Select Record
                const Selected = await this.notResetable().select([], throwError);
                // Update Record
                await this.isResetable().update(data, throwError);
                return Selected;
            }
            catch (error) {
                this.isResetable().reset();
                throw error;
            }
        };
        this.selectOneAndUpdate = async (data, throwError = false) => {
            try {
                // Select Record
                const Selected = await this.notResetable().selectOne([], throwError);
                // Update Record
                await this.isResetable().update(data, throwError);
                return Selected;
            }
            catch (error) {
                this.isResetable().reset();
                throw error;
            }
        };
        this.updateAndSelect = async (data, throwError = false) => {
            try {
                // Update Record
                await this.notResetable().update(data, throwError);
                // Select Record
                return this.isResetable().select([], throwError);
            }
            catch (error) {
                this.isResetable().reset();
                throw error;
            }
        };
        this.updateAndSelectOne = async (data, throwError = false) => (await this.updateAndSelect(data, throwError))[0];
        this.delete = async (throwError = false) => {
            try {
                // Get Conditions
                const QueryConditions = this.ConditionsBuilder.render(false);
                // Delete Records
                const Result = (await this.query(`DELETE FROM \`${this.getSchemaName()}\` ${QueryConditions.sql};`, QueryConditions.values));
                // Reset Schema
                this.reset();
                // Resolve Response
                if (Result &&
                    Result instanceof Array &&
                    Result[0] &&
                    typeof Result[0].affectedRows !== "undefined") {
                    if (throwError && !Result[0].affectedRows)
                        throw new Error("Nothing has been deleted!");
                    else
                        return Result[0];
                }
                else
                    throw new Error("Unknown results encountered!");
            }
            catch (error) {
                this.reset();
                throw error;
            }
        };
        this.deleteOrFail = () => this.delete(true);
        this.selectAndDelete = async (throwError = false) => {
            try {
                // Select Record
                const Selected = this.notResetable().select([], throwError);
                // Delete Record
                this.isResetable().delete(throwError);
                return Selected;
            }
            catch (error) {
                this.isResetable().reset();
                throw error;
            }
        };
        this.selectOneAndDelete = async (throwError = false) => (await this.selectAndDelete(throwError))[0];
        this.empty = async () => {
            const Result = (await this.reset().query(`TRUNCATE TABLE \`${this.getSchemaName()}\`;`));
            if (Result &&
                Result instanceof Array &&
                Result[0] &&
                typeof Result[0].affectedRows !== "undefined") {
                return Result[0];
            }
            else
                throw new Error("Unknown results encountered!");
        };
        this.isResetable = () => {
            this.Resetable = true;
            return this;
        };
        this.notResetable = () => {
            this.Resetable = false;
            return this;
        };
        this.reset = () => {
            if (this.Resetable) {
                // Reset Relations
                this.resetRelation();
                // Reset Default Selection
                this.resetSelection();
                // Reset Default Columns
                this.resetColumns();
                // Reset Conditions Builder
                this.ConditionsBuilder.reset();
            }
            return this;
        };
        this.Type = this.Schema.Types.length ? this.Schema.Types[0] : "";
        this.ConnectionManager = this.Schema.ConnectionManagers[this.Id];
    }
}
exports.LockedSchema = LockedSchema;
