import { ConnectionManager } from "./connection-manager";
import { SQLBoolean } from "./data-types/boolean";
import { SQLContent } from "./data-types/content";
import { SQLEnum } from "./data-types/enum";
import { SQLManyRelation } from "./data-types/many-relation";
import { SQLMeta } from "./data-types/meta";
import { SQLNumber } from "./data-types/number";
import { SQLObject } from "./data-types/object";
import { SQLOneRelation } from "./data-types/one-relation";
import { SQLString } from "./data-types/string";
import { LockedSchema } from "./locked-schema";
export declare type PhysicalColumnTypes = SQLString<any, any, any> | SQLContent<any, any, any> | SQLEnum<any, any, any> | SQLNumber<any, any, any> | SQLBoolean<any, any, any> | SQLObject<any, any, any>;
export declare type VirtualColumnTypes = SQLOneRelation<any> | SQLManyRelation<any> | SQLMeta<any>;
declare type SQL_SCHEMA_ALLOWED_TYPES = PhysicalColumnTypes | VirtualColumnTypes;
declare type SCHEMA_RESERVED = {
    _id?: never;
    _headers?: never;
};
declare type Indexes = "INDEX" | "UNIQUE" | "FULLTEXT" | "SPATIAL";
export declare type SQL_SCHEMA = {
    [key: string]: SQL_SCHEMA_ALLOWED_TYPES;
} & SCHEMA_RESERVED;
export declare type SchemaIndex<T extends SQL_SCHEMA> = {
    name: string;
    index: Indexes;
    columns: Array<keyof T>;
};
export declare type RenderedSchema<T extends Schema<any>> = {
    name: string;
    schema: T;
};
export declare class Schema<T extends SQL_SCHEMA> {
    Name: string;
    Data: T;
    Prefix?: string;
    Types: string[];
    Indexes: Array<SchemaIndex<T>>;
    SearchableColumns: (keyof T)[];
    metaResolvers: Record<string, SQLMeta<any>>;
    ConnectionManagers: {
        [Key: string]: ConnectionManager<any>;
    };
    constructor(Name: string, Data: T);
    setConnection: (id: string, ConnectionManager: ConnectionManager<any>) => this;
    hasConnection: (id?: string | undefined) => boolean;
    deleteConnection: (id: string) => boolean;
    Render: () => RenderedSchema<Schema<T>>[];
    addPrefix: (prefix: string) => this;
    addType: (type: string) => this;
    addIndex: (index: Indexes, name: string, columns: Array<keyof T>) => this;
    columns: () => (keyof T & string)[];
    hasColumn: (column: string) => boolean;
    getColumn: (column: string) => SQL_SCHEMA_ALLOWED_TYPES;
    physicalColumnNames: () => string[];
    virtualColumnNames: () => string[];
    physicalColumns: () => PhysicalColumnTypes[];
    virtualColumns: () => VirtualColumnTypes[];
    hasPhysicalColumn: (column: string) => boolean;
    hasVirtualColumn: (column: string) => boolean;
    lock: (id: string) => LockedSchema<this>;
    new: (id: string) => LockedSchema<this>;
}
export {};
