"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Schema = void 0;
const connection_manager_1 = require("./connection-manager");
const meta_1 = require("./data-types/meta");
const primaryKey_1 = require("./data-types/primaryKey");
const exceptions_1 = require("./exceptions");
const locked_schema_1 = require("./locked-schema");
class Schema {
    constructor(Name, Data) {
        this.Name = Name;
        this.Data = Data;
        this.Types = [];
        this.Indexes = [];
        this.SearchableColumns = [];
        this.metaResolvers = {};
        this.ConnectionManagers = {};
        this.setConnection = (id, ConnectionManager) => {
            this.ConnectionManagers[id] = ConnectionManager;
            return this;
        };
        this.hasConnection = (id) => id
            ? this.ConnectionManagers[id] instanceof connection_manager_1.ConnectionManager
            : !!Object.keys(this.ConnectionManagers).length;
        this.deleteConnection = (id) => delete this.ConnectionManagers[id];
        this.Render = () => (this.Types.length ? this.Types : [""]).map((type) => ({
            name: [this.Prefix, this.Name, type].filter((v) => v).join("-"),
            schema: this,
        }));
        this.addPrefix = (prefix) => {
            if (typeof prefix === "string")
                this.Prefix = prefix;
            return this;
        };
        this.addType = (type) => {
            if (typeof type === "string")
                this.Types.push(type);
            return this;
        };
        this.addIndex = (index, name, columns) => {
            this.Indexes.push({ name, index, columns });
            return this;
        };
        this.columns = () => Object.keys(this.Data);
        this.hasColumn = (column) => this.columns().includes(column);
        this.getColumn = (column) => this.Data[column];
        this.physicalColumnNames = () => Object.keys(this.Data).filter((columnKey) => !this.Data[columnKey].isVirtual);
        this.virtualColumnNames = () => Object.keys(this.Data).filter((columnKey) => this.Data[columnKey].isVirtual);
        this.physicalColumns = () => this.physicalColumnNames().map((columnKey) => this.Data[columnKey]);
        this.virtualColumns = () => this.virtualColumnNames().map((columnKey) => this.Data[columnKey]);
        this.hasPhysicalColumn = (column) => this.physicalColumnNames().includes(column);
        this.hasVirtualColumn = (column) => this.virtualColumnNames().includes(column);
        this.lock = (id) => new locked_schema_1.LockedSchema(id, this);
        this.new = (id) => this.lock(id);
        // Schema Validation
        if (this.Data._id)
            throw new exceptions_1.SQLException(`Column with name '_id' is reserved! Please try different column name.`);
        else if (this.Data._headers)
            throw new exceptions_1.SQLException(`Column with name '_headers' is reserved! Please try different column name.`);
        // Add _id To Schema
        this.Data = Object.assign({ _id: new primaryKey_1.PrimaryKey() }, this.Data);
        // Add Column IDs
        Object.keys(this.Data).forEach((columnId) => (this.Data[columnId].columnId = columnId));
        // Collect Meta Columns
        this.virtualColumns().forEach((column) => {
            if (column instanceof meta_1.SQLMeta)
                this.metaResolvers[column.columnId] = column;
        });
        // Collect Searchable Columns
        this.SearchableColumns = this.physicalColumns()
            .map((column) => column.isSearchable && !column.isBinary ? column.columnId : undefined)
            .filter((v) => v);
        // Create Full Text Search Index
        if (this.SearchableColumns.length)
            this.addIndex("FULLTEXT", "_search", this.SearchableColumns);
    }
}
exports.Schema = Schema;
